# -*- coding: utf-8 -*-
"""
Created on Feb 09 2016 09:37

@author: Jan.Reimes
"""

from collections import OrderedDict
import pandas
import numpy as np
from scipy.stats import pearsonr, spearmanr, kendalltau
from common.math.polyFitConstrained import polyfit_constr

def rmse(x,y,d):
    # according to P.1401, 7.5.1.
    N = np.max(x.shape)
    return np.sqrt(np.sum(np.power(x-y, 2)) / (N-1-d))

class Performance:
    strErrRaw = 'Raw Error'
    strErrMapped = 'Error after mapping'
    strErrDStar = 'Error after DeltaStar'
    strErrMappedDStar = 'Error after mapping and subsequent DeltaStar'
    #strErrDStarMapped = 'Error after DeltaStar and subsequent mapping'

    def __init__(self, title, mapping='poly', degree=3, minValue=1.0, maxValue=5.0, **kwargs):
        self.mapping = mapping
        self.degree = degree
        self.minValue = minValue
        self.maxValue = maxValue
        self.title = title
        self.constr_weight_mean = kwargs.get("constr_weight_mean", 1.0) # constraint only wrt monotonicity

    def deltaStar(self, subj, obj, ci95):
        # correct subjective data towards objective
        delta = subj - obj
        delta_star = (np.abs(delta) - ci95) * np.sign(delta)
        delta_star[np.abs(delta) <= ci95] = 0
        subj_star = np.clip(obj + delta_star, a_min=self.minValue, a_max=self.maxValue)
        return subj_star

    def applyMapping(self, subj, obj):
        # map objective towards subjective
        if self.mapping == 'poly':
            fun = np.poly1d(np.polyfit(obj, subj, self.degree))
        elif self.mapping == 'polyc':
            fun = polyfit_constr(obj, subj, self.degree, xmin=self.minValue, xmax=self.maxValue, weight_mean=self.constr_weight_mean)
        else:
            # To Do: Add more, e.g. sigmoid mapping...
            raise Exception("Invalid mapping type!")

        obj_mapped = pandas.Series(index=obj.index, data=fun(obj))
        return obj_mapped

    def getMetrics(self, x, y, valuesAreMapped):
        if valuesAreMapped:
            d = self.degree
        else:
            d = 0

        result = OrderedDict()
        result['rmse'] = rmse(x,y, d)
        result['corr'] = pearsonr(x,y)[0]
        result['spear'] = spearmanr(x,y)[0]
        result['kendall'] = kendalltau(x,y)[0]
        result['maxabs'] = np.max(np.abs(x-y))
        return result

    def calculateMetrics(self, subj, obj, ci95):
        # check input variables, convert everything to Series
        if not type(subj) is pandas.Series:
            subj = pandas.Series(data=subj)

        if not type(obj) is pandas.Series:
            obj = pandas.Series(data=obj)

        # default if no ci95 is provided: CI95 == 0.0
        if ci95 is None:
            ci95 = pandas.Series(index=subj.index, data=np.zeros_like(subj))

        df = OrderedDict()

        # generate different versions of subj and obj by mapping and deltastar...
        # objective scores mapping towards subjective:
        obj_mapped = self.applyMapping(subj, obj)
        # subjective scores corrected according to CI95 towards objective
        subj_star = self.deltaStar(subj, obj, ci95)
        # subjective scores corrected according to CI95 towards mapped objective
        subj_star_on_mapped = self.deltaStar(subj, obj_mapped, ci95)
        # experimental: objective scores mapped towards subjective corrected scores
        #obj_mapped_on_star = self.applyMapping(subj_star, obj)

        # calculate error signals for evaluation purposes
        error_raw = subj - obj
        error_raw.name = self.strErrRaw

        error_mapped = subj - obj_mapped
        error_mapped.name = self.strErrMapped

        error_star = subj_star - obj
        error_star.name = self.strErrDStar

        error_mapped_star = subj_star_on_mapped - obj_mapped
        error_mapped_star.name = self.strErrMappedDStar

        #error_star_mapped = subj_star - obj_mapped_on_star
        #error_star_mapped.name = self.strErrDStarMapped
        error_star_mapped = None

        # raw error
        df['raw'] = self.getMetrics(subj, obj, False)

        # mapped
        df['mapped'] = self.getMetrics(subj, obj_mapped, True)

        # raw error incl. star-operator
        df['dstar'] = self.getMetrics(subj_star, obj, False)

        # mapped + star => RMSE* acc. to ITU-T P.1401!
        df['mapped+dstar'] = self.getMetrics(subj_star_on_mapped, obj_mapped, True)

        # star + mapped - experimental; could be removed?
        #df['dstar+mapped'] = self.getMetrics(subj_star, obj_mapped_on_star, True)

        # to data frame
        df = pandas.DataFrame(df)
        df.index.name = 'Metric'
        df['Title'] = self.title
        df.set_index('Title', append=True, inplace=True)
        df = df.reorder_levels(['Title', 'Metric'])

        return df, error_raw, error_mapped, error_star, error_mapped_star, error_star_mapped

if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from common.math.GenerateTestData import createTestData

    MOS_Subj, CI95, MOS_Obj, N = createTestData()

    ### test performance metrics: ###
    degree = 3
    df, error_raw, error_mapped, error_star, error_mapped_star, error_star_mapped = Performance('Database#01 - Default', degree=degree).calculateMetrics(MOS_Subj, MOS_Obj, CI95)
    print(df)
    df, error_raw, error_mapped, error_star, error_mapped_star, error_star_mapped = Performance('Database#01 - Constrained', degree=degree, mapping='polyc', constr_weight_mean=0.15).calculateMetrics(MOS_Subj, MOS_Obj, CI95)
    print(df)

    #df.to_excel('TestData.xlsx')

    ### plot ###
    pol_c = polyfit_constr(MOS_Obj, MOS_Subj, degree, xmin=minMOS, xmax=maxMOS, weight_mean=1, fixed_points=[(5,5)], fixed_points_distance=[0.25])
    pol = np.poly1d(np.polyfit(MOS_Obj, MOS_Subj, degree))

    xx = np.linspace(minMOS, maxMOS, num=N*2)
    yy = pol(xx)
    yy_c = pol_c(xx)

    plt.scatter(MOS_Obj, MOS_Subj, label=mosTitle)
    plt.scatter(MOS_Obj[I], MOS_Subj[I], label='max abs delta')
    plt.plot(xx, yy, 'r', label='default')
    plt.plot(xx, yy_c, 'b--', label='constrained')
    plt.grid(True)
    plt.legend()
    plt.xlim(np.round([minMOS, maxMOS]))
    plt.ylim(np.round([minMOS, maxMOS]))
    plt.xlabel('Instrumental %s' % mosTitle)
    plt.ylabel('Auditory %s' % mosTitle)
    plt.show()