# -*- coding: utf-8 -*-
"""
Created on Feb 20 2017 22:23

@author: Jan.Reimes
"""

import numpy as np
from scipy.optimize import minimize

class PolyFitMeanMax:
    def __init__(self, degree=3, weight_mean=0.5, **kwargs):
        self.weight_mean = weight_mean
        self.degree = degree
        self.constraints = []

    # polynom of function
    def polynomial(self, p, _x):
        if len(_x.shape) == 1:
            _x = _x[:, np.newaxis]

        _y = np.dot(_x ** np.arange(self.degree, -1, step=-1), p)
        return _y

    # function to minimize
    def objective(self, p, _x, _y):
        error = (self.polynomial(p, _x) - _y) ** 2
        return self.weight_mean * error.mean() + (1-self.weight_mean) * error.max()

    def fit(self, x, y, **kwargs):
        # initial guess: y = x
        x0 = np.zeros((self.degree + 1))
        x0[-2] = 1.0

        # do minimization!
        res = minimize(self.objective, x0, args=(x, y), constraints=self.constraints)
        if not res.success:
            print("Warning: %s [%d]" % (res.message, res.status))

        # return polynom object
        return np.poly1d(res.x)

class PolyFitMonotonicity(PolyFitMeanMax):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xmin = kwargs.get('xmin', None)
        self.xmax = kwargs.get('xmax', None)
        self.nbrSteps = kwargs.get('nbrSteps', 50)
        # penalty value for 1st derivate
        self.__min_dy = kwargs.get('min_dy', -1e-3)

    # polynom of 1st derivate
    def constraint_1st_deriv(self, p, _xrange):
        deriv = 1

        _x = _xrange
        if len(_xrange.shape) == 1:
            _x = _xrange[:, np.newaxis]

        _yd = np.dot(_x ** np.arange(self.degree-deriv, -1, step=-1), p[:-deriv])
        return _yd.min() + self.__min_dy

    def fit(self, x, y, **kwargs):
        # default values:
        if not self.xmin:
            self.xmin = x.min()

        if not self.xmax:
            self.xmax = x.max()

        # set constraint: monotonity
        xrange = np.linspace(self.xmin, self.xmax, num=self.nbrSteps)
        self.constraints.append({'type': 'ineq', 'fun': self.constraint_1st_deriv, 'args': (xrange,)})

        # perform optimization
        return super().fit(x,y, **kwargs)

class PolyFitFixedPoints(PolyFitMeanMax):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # list of tuples e.g. (x=5,y=5)
        self.fixed_points = kwargs.get('fixed_points', [])
        # list of allowed distance, e.g. [0.1]
        self.fixed_points_distance = kwargs.get('fixed_points_distance', [])

    def constraint_fixed_point(self, p, point, allowed_tolerance):
        xp, yp = point
        #abs_error = np.maximum(np.abs(self.polynomial(p, np.array([xp]))[0] - yp) - allowed_tolerance, 0.0)
        abs_error = np.abs(self.polynomial(p, np.array([xp]))[0] - yp)
        return abs_error**1

    def fit(self, x, y, **kwargs):
        # set constraint: fixed points and allowed tolerance
        if (len(self.fixed_points_distance) != len(self.fixed_points)) and len(self.fixed_points_distance) > 0:
            raise Exception("fixed_points_distance must be empty or same size as fixed_points!")

        for i, fpoint in enumerate(self.fixed_points):
            allowed_tolerance = 0
            if len(self.fixed_points_distance) > 0:
                allowed_tolerance = self.fixed_points_distance[i]
            self.constraints.append({'type': 'eq', 'fun': self.constraint_fixed_point, 'args': (fpoint, allowed_tolerance)})

        # perform optimization
        return super().fit(x,y, **kwargs)

# Combination of both - which order?
class PolyFitFixedPointsMonotonicity(PolyFitFixedPoints, PolyFitMonotonicity):
    pass

class PolyFitMonotonicityFixedPoints(PolyFitMonotonicity, PolyFitFixedPoints):
    pass


def polyfit_constr(x, y, deg, xmin=None, xmax=None, nbrSteps=50, weight_mean=0.5, fixed_points=[], fixed_points_distance=[]):
    #theClass = PolyFitFixedPointsMonotonicity
    #theClass = PolyFitMonotonicityFixedPoints
    theClass = PolyFitMonotonicity
    obj = theClass(degree=deg, xmin=xmin, xmax=xmax, nbrSteps=nbrSteps, weight_mean=weight_mean, fixed_points=fixed_points, fixed_points_distance=fixed_points_distance)
    return obj.fit(x,y)

if __name__ == "__main__":
    # removed demo code here...
    pass