# -*- coding: utf-8 -*-
"""
Created on Wed Aug 27 18:01:27 2014

@author: Jan.Reimes
"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import pearsonr, spearmanr, kendalltau
from sklearn.metrics import mean_squared_error
from math import sqrt
from PIL import Image
from common.math.polyFitConstrained import polyfit_constr
from common.math.PerformanceMetrics import Performance

### The plotting function ###
def PlotCorr(dataX, dataY, axLabels=['Instrumental MOS', 'Auditory MOS'], MappingOrder=3, xRange=[1,5], yRange=[1,5], legendText="MOS", title="Results", **kwargs):
    ########################################################
    # Consts for drawing
    lw = 2.5 # line width
    nbrPointsMappingFunction = 50
    defaultFontSize = 12
    defaultScatterDotSize = 50
    ########################################################
    # General options:
    ShowTubeLines = kwargs.get("ShowTubeLines", True)
    TubeLinesDistance = kwargs.get("TubeLinesDistance", 0.5)
    xErr = kwargs.get('xErr', None)
    yErr = kwargs.get('yErr', None)

    # Metric options:
    ShowCorrelationInLegend = kwargs.get('ShowCorrelationInLegend', False)
    ShowPolynomCoeffs = kwargs.get('ShowPolynomCoeffs', True)
    MetricsToDisplay = kwargs.get("MetricsToDisplay", ["rmse", "maxabs"] ) # ["corr", "spear", "kendall", "rmse"]
    MetricTypes = kwargs.get("MetricTypes", ['mapped+dstar']) # ['raw', 'mapped', 'dstar', 'mapped+dstar']

    # Mapping options:
    ShowMappedValues = kwargs.get("ShowMappedValues", False)
    ShowMappingFunction = MappingOrder >= 1
    ConstrainedMapping = kwargs.get("ConstrainedMapping", True)
    ConstrainedMappingMeanWeight = kwargs.get("ConstrainedMappingMeanWeight", 1.0)

    # Plot Options:
    scatterDotSize = kwargs.get("scatterDotSize", defaultScatterDotSize)
    fontsize = kwargs.get("fontsize", defaultFontSize)
    ########################################################
    # Input data - convert to NumPy
    x = np.array(dataX)
    y = np.array(dataY)

    ########################################################
    # Mapping functions
    fx = np.linspace(start=xRange[0], stop=xRange[1], num=nbrPointsMappingFunction)
    fy = np.linspace(start=yRange[0], stop=yRange[1], num=nbrPointsMappingFunction)
    if ConstrainedMapping:
        # constrained mapping regarding mean/max and monotonicity optimization
        p = polyfit_constr(x, y, MappingOrder, xmin=xRange[0], xmax=xRange[1], weight_mean=ConstrainedMappingMeanWeight)
    else:
        # default polyfit(),
        p_fit = np.polyfit(x, y, MappingOrder)
        p = np.poly1d(p_fit)

    ########################################################
    # Correlation coeffs, error metrics
    mapping = 'polyc' if ConstrainedMapping else 'poly'
    perf = Performance(str(title), mapping=mapping, constr_weight_mean=ConstrainedMappingMeanWeight)
    df = perf.calculateMetrics(y, x, yErr)[0]
    df.index = df.index.droplevel(0)

    r_p = df.loc['corr', 'raw']
    r_p_m = df.loc['corr', 'mapped']
    ########################################################
    # Start figure
    fig, ax = plt.subplots(1,1, figsize=(4.8, 4.8), dpi=100)
    ax.set_aspect('equal')

    ax.plot(fx, fy, 'k--', linewidth=lw)
    if ShowMappedValues:
        label = legendText if not ShowCorrelationInLegend else "%s (r=%4.3f)" % (legendText, r_p_m)
        scatter = ax.scatter(p(x),y, s=scatterDotSize, label=label)
    else:
        label = legendText if not ShowCorrelationInLegend else "%s (r=%4.3f)" % (legendText, r_p)
        scatter = ax.scatter(x,y, s=scatterDotSize, label=label)
        if ShowMappingFunction:
            label = 'Mapping function (r=%4.3f)' % (r_p_m) if ShowCorrelationInLegend else 'Mapping function'
            ax.plot(fx, p(fx), 'r--', label=label, linewidth=lw)

    if (not xErr is None) or (not yErr is None):
        if ShowMappedValues and (not xErr is None):
            # transform yerr:
            xUpperTr = p(x + xErr) - p(x)
            xLowerTr = np.abs(p(x - xErr) - p(x))
            ax.errorbar(p(x), y, xerr=np.array([xLowerTr, xUpperTr]), yerr=yErr, fmt='bo')
        else:
            ax.errorbar(x,y, xerr=xErr, yerr=yErr, fmt='o')

    ax.grid(True)
    ax.set_xlabel(axLabels[0])
    ax.set_ylabel(axLabels[1])
    ax.set_xlim(xRange)
    ax.set_ylim(yRange)

    if ShowTubeLines:
        # nur die Eck-Punkte plotten
        xx = np.array(xRange)
        yy = np.array(yRange)
        ax.plot(xx,yy, 'k--')
        ax.plot(xx,yy+TubeLinesDistance, 'g--')
        ax.plot(xx,yy-TubeLinesDistance, 'g--')

    ########################################################
    # draw metrics in upper left corner
    deltaX = xRange[1]-xRange[0]
    deltaY = yRange[1]-yRange[0]
    startPos = yRange[1] - (deltaY*0.05)
    offset_y = -(deltaY)/15
    offset_x = xRange[0] + (deltaX*0.05)
    currentPosY = startPos

    # coeffs
    if ShowPolynomCoeffs:
        currentPosY += offset_y/4
        str_poly = str(p).split("\n")
        str_poly[1] = "y = " + str_poly[1]
        str_poly[0] = " "*4 + str_poly[0]
        str_poly = "\n".join(str_poly)
        ax.text(offset_x, currentPosY, str_poly, family='monospace', size='x-small')
        currentPosY += offset_y

    for metric in MetricsToDisplay:
        for metricType in MetricTypes:
            value = df.loc[metric, metricType]
            dstar = '*' if ('dstar' in metricType) else ''
            mapped = ' (mapped)' if ('mapped' in metricType) else ''
            ax.text(offset_x,currentPosY, '%s%s%s = %.3f' % (metric, dstar, mapped, value))
            currentPosY += offset_y


    ########################################################
    # scatter on top
    scatter.set_zorder(100)

    # highlight maxabs?
    if 'maxabs' in MetricsToDisplay:

        if ShowMappedValues:
            I = np.argmax(np.abs(y - p(x)))
            scatter1 = ax.scatter(p(x[I]),y[I], color='orange', s=scatterDotSize  )
        else:
            I = np.argmax(np.abs(y - x))
            scatter1 = ax.scatter(x[I],y[I], color='orange', s=scatterDotSize)
        scatter1.set_zorder(101)
    ########################################################
    # Info
    plt.legend(loc='lower right')
    if title:
        ax.set_title(title)

    ########################################################
    # Font Size
    for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()):
         item.set_fontsize(fontsize)


    #fig.subplots_adjust(top=0.05)
    fig.tight_layout()
    return fig, ax
########################################################################
def MergeFiles(SrcFiles, OutputFile, Spacing=0, horizontal=True, defaultColor=16777215):
    images = []
    NewSizeX = 0
    NewSizeY = 0
    for srcFile in SrcFiles:
        im = Image.open(srcFile)
        images.append(im)
        if horizontal:
            NewSizeX += im.size[0] + Spacing
            NewSizeY = np.max((im.size[1], NewSizeY))
        else:
            NewSizeX = np.max((im.size[0], NewSizeX))
            NewSizeY += im.size[1] + Spacing

    # einmal am ende wieder abziehen
    if horizontal:
        NewSizeX -= Spacing
    else:
        NewSizeY -= Spacing

    new = Image.new('RGB', (NewSizeX, NewSizeY), color=defaultColor)

    newPosX = 0
    newPosY = 0
    lastImage = None
    for i, im in enumerate(images):
        if i > 0:
            if horizontal:
                newPosX += lastImage.size[0] + Spacing
            else:
                newPosY += lastImage.size[1] + Spacing
        new.paste(im, (newPosX,newPosY))
        lastImage = im

    new.save(OutputFile)


########################################################################
if __name__ == "__main__":
    pass